module UU
  module OS
    module Util
      module ResponseBuilder

        # Method for construction of custom response.
        # @param [Fixnum] status HTTP status code (see {http://en.wikipedia.org/wiki/List_of_HTTP_status_codes
        #    List of HTTP status codes} for details).
        # @param [Hash, UU::OS::HTTP::HTTPHeaders] headers Response headers (multi-value headers should1
        #    be given as array value).
        # @param [String, IO] body Response body.
        # @return [Object] Response object to be handled by server.
        def response(status, headers, body)
          if !headers.nil? && !headers.kind_of?(Hash)
            raise "invalid headers, expected hash but was #{headers.class}"
          end
          if !body.nil? && !body.kind_of?(String) && !body.respond_to?(:read)
            raise "cannot send #{body.class} as response"
          end
          headers ||= {}
          headers = headers.inject({}) do |tmp, (key, value)|
            key = key.to_s.gsub('_', '-').downcase
            if value.kind_of?(Array)
              tmp[key] = value.join(',')
            else
              tmp[key] = value.to_s
            end
            tmp
          end
          headers['content-type'] ||= 'application/octet-stream'

          body = body || ''
          if body.respond_to?(:read)
            body = ChunkedIO.new(body)
          elsif !body.respond_to?(:each)
            body = [body]
          end
          # Actual support of streaming depends on particular implementation of
          # rack handler. Body is read using each, however handler implementation
          # provide output into which body is written.
          return [status.to_i, headers, body]
        end

        # Constructs file download response with HTTP status code 200
        # (should be used to return binary data).
        # @param [String, IO] body Object representing binary content.
        # @param [Hash] meta_data Additional optional response attributes
        #    (if not set, they are obtained from body if body provides them):
        # @option meta_data [String] :name File name
        # @option meta_data [String] :size Response content size
        # @option meta_data [String] :content_type Content type
        # @param [Proc] block Optional block with IO parameter for asynchronous
        #    streaming of binary content (if given, +body+ attribute is ignored).
        # @return [Object] Response object to be handled by server.
        def file_download_response(body = nil, meta_data = nil, &block)
          body, meta_data = meta_data, body if body.kind_of?(Hash)
          if !body.nil? && !body.kind_of?(String) && !body.respond_to?(:read)
            raise "cannot send #{body.class} as response"
          end
          meta_data ||= {}
          meta_data[:name] ||= File.basename(body) if body.respond_to?(:path)
          meta_data[:size] ||= body.size if body.respond_to?(:size)
          meta_data[:content_type] ||= body.content_type if body.respond_to?(:content_type)
          headers = {}
          headers[:content_type] = meta_data[:content_type] || 'application/octet-stream'
          headers[:content_disposition] = 'attachment'
          headers[:content_disposition] << "; filename=\"#{meta_data[:name]}\"" if meta_data[:name]
          headers[:content_length] = meta_data[:size] if meta_data[:size]
          if block.nil?
            return response(200, headers, body)
          else
            return streamed_response(headers, &block)
          end
        end

        # Method allows asynchronous streaming of body with HTTP status code 200.
        # @param [Hash, UU::OS::HTTP::HTTPHeaders] headers Response headers
        #    (multi-value headers should be given as array value).
        # @param [Proc] block Block with IO parameter for asynchronous streaming.
        # @return [Object] Response object to be handled by server.
        def streamed_response(headers, &block)
          raise 'no block' if block.nil?
          read, write = IO.pipe
          write.write('')
          Thread.new {
            block.yield(write)
          }
          return response(200, headers, read)
        end

        # Class allowing chunked streaming of IO response body (by default
        # each on IO is in fact "each_line" and therefore is not suitable
        # for binary data streaming).
        # @private
        class ChunkedIO # :nodoc:
          def initialize(io)
            @io = io
          end

          def each(&block)
            while (!(chunk = @io.read(65535)).nil?)
              block.yield(chunk)
            end
            @io.close if @io.respond_to?(:close)
          end
        end

      end
    end
  end
end
